classdef MerbonDatabase

    properties(GetAccess = public, SetAccess = private)
        % credentials for web service
        password
        username
        endpointUrl
    end
    properties(Access = private)
        verboseMode = true;  
        valOffset = 0;
        varOffset = 0;
        valCount = 10000;
        varCount = 10000;
        dbDouble = 0;
        dbInt64 = 3;
        dbBool = 6;
        dbString = 2;
    end
    
    methods
        function this = MerbonDatabase(username,password,endpointUrl,verboseMode)
            % Create an object of MerbornDatabase with these properties:
            %
            % INPUTS:
            % username - username to access the DB.
            % password - password to access the DB.
            % endpointUrl - address of database.
            % verboseMode - on/off verbose mode. true - on, false - off.
            %
            % OUTPUTS:
            % MerbonDatabase object;
            %
            % call Example: 
            % db = MerbornDatabase('user_name','password','endpointUrl',verboseMode)
            % db = MerbornDatabase('user_name','password','endpointUrl')
            
            if (nargin < 4)
                this.verboseMode = true;
            else
                this.verboseMode = verboseMode;
            end  
            if (nargin < 3)
                 error('RcWare client initialization failed - not enough arguments');
            end
%                 msg = sprintf('Creating object of MerbonDatabase was succesfull.\nVerbose mode was automatically set up on true value');
%                 DisplayProcess(this.verboseMode,msg,0,0);

            this.username = username;
            this.password = password;
            this.endpointUrl = endpointUrl;
        end
        
        
        function MerbonDbVariables = getAllVars(this,varDef)
            % Download all variables from DB, that match to the
            % input key definition. If no key definition is put
            % then will be downloaded all varibles from entire DB.
            %
            % INPUTS:
            % varDef - cell array of varible keys
            %
            % OUTPUTs:
            % MerbonDbVariables - array of MerbonDbVariable objects. Is downloaded
            % as much variables as much matches are fount among keys at
            % input and in DB.
            %
            % call example:
            % varDef1 = {'StationName','ENERGO';'DPName','venkovni teplota'};
            % varDef2 = {'StationName','ENERGO';'DPName','vnitrni teplota'}; 
            % allVariables = db.getAllVars([varDef; varDef2]);
            % allVariables = db.getAllVars();
            
            if nargin < 2
                warning('Input argumetn varDef is missing. All data will be dowlanded from the database.\nDo you wish to continue? y/n');
                str = input('y/n ---> ', 's'); 
                
                if(strcmp(str,'y'))
%                      msg = sprintf('Dowlanding all Variables.\nDots mean that dowlanding is still in process.');
%                      DisplayProcess(this.verboseMode,msg,0,0);
                     tic;
                     clear mex;
                     vars = MexMain({this.username, this.password, this.endpointUrl}, 100, {this.varOffset, this.varCount });
                     MerbonDbVariables = MerbonDbVariable.empty(length(vars),0);      
                     for i=1:length(vars)
                        actVar = vars{i};
                        MerbonDbVariables(i) = MerbonDbVariable(actVar);
                     end
                     t= toc;
                     msg = sprintf('\nAll variables were successfully downloaded . (%d variables/%.2f sec)',length(MerbonDbVariables),t);
                     DisplayProcess(this.verboseMode,msg,0,0);
                elseif(strcmp(str,'n'))
                     return;
                else
                     error('Undefined operation');
                end
            else
%             msg = sprintf('Dowlanding defined Variables.\nDots mean that dowlanding is still in process.');
%             DisplayProcess(this.verboseMode,msg,0,0);
            tic;
            clear mex;
            vars = MexMain({this.username, this.password, this.endpointUrl}, 0, varDef, {this.varOffset, this.varCount });
            MerbonDbVariables = MerbonDbVariable.empty(length(vars),0);      
            for i=1:length(vars)
               actVar = vars{i};
               MerbonDbVariables(i) = MerbonDbVariable(actVar);
            end
               t= toc;
               msg = sprintf('\nAll variables were successfully downloaded. (%d variables/%.2f sec)',length(MerbonDbVariables),t);
               DisplayProcess(this.verboseMode,msg,0,0);
            end           
        end
        
        function [data,time,dbVars,dataInCell,anyErrors,isCharVariable,strMaps,dataInfo] = getDataUnmerged(this,dbVars,from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings,stringRegime)
            % Download data from database.
            %
            % INPUTS:
            % dbVars  - array of MerbonDbVarible objects
            % from - matlab format time in utc. Downloading will start at this time
            % to - matlab format time in utc. Dowlanding will end at this time
            % samplingTime - sampling period in the specified time interval. In seconds
            % convertToLocal - Define output time kind. Possible two parameters: 'Local' or 'Utc' - default
            % interpolationMethod - interpolation method (optional):
            % 'nearest','spline','previous','linear' - default.
            % displayWarnings - true/false. Display warnings which occurs during process (optional)
            % stringRegime = true/false. Define if string varibales will be
            % represented as strings or via theirs hash code (optional).
            %
            % OUTPUS:
            % data - matrix of downloaded data. Each column represent one
            % varibale. Each row represent value in sample.
            % time - vector of time stamps
            % dbVars - array of MerbonDbVariable objects. Has the same
            % lengths as number of columns in data matrix. Data at n-th
            % column in data matrix correspond with variable definition at n-th position
            % in dbVars.
            % dataInCell - cell array of downloaded data. String are
            % represented as strings or hash codes - dependeds on
            % stringRegime input.
            % anyErrors - flag if any errors occurs during process.
            % isCharVariable - boolean vector which indicates if n-th
            % varible is string-type or not.
            % strMaps - dictionary of string hash codes and theirs string
            % interpretations. Is returned if stringRegime is set to false
            % dataInfo - array of structures which keeps additional information about
            % downloaded variables. N-th structure correspond with n-th
            % variable.
            %
            % call example:
            % dbVars = MerbonDbVariable({'GroupName','Energo'});
            % from = datetime(2017, 1, 1);
            % to = datetime(2017, 31, 1); 
            % [data,time,vars, dataInCells] = getDataUnmerged(dbVars, from, to, 5*60, 'Local', 'Linear', false)
            % [data,time,vars, dataInCells, ~, charVariables, straMap, dataInfo] = getDataUnmerged(dbVars, from ,to ,15*60,'Utc','linear',true, false);
            
            clear mex;
            anyErrors = false;
            
            % inuputs checking
            if(isnumeric(from) && isnumeric(to))
                timeTo = matlab2unix(to);
                timeFrom = matlab2unix(from);
            else
                error('Time must be inserted as numeric Matlab format');
            end
            
            if nargin < 5
                convertToLocal = 'utc';
                samplingTime = 5*60;
                displayWarnings = true;
                stringRegime = true;
            end
            if nargin < 6
                convertToLocal = 'utc';
                interpolationMethod = 'linear';
                displayWarnings = true;
                stringRegime = true;
            end
            if nargin < 7
                interpolationMethod = 'linear';
                displayWarnings = true;
                stringRegime = true;
            end
            if nargin < 8 
                displayWarnings = true;
                stringRegime = true;
            end
            if(nargin < 9)
                stringRegime = true;
            end
             
            % process
            time = timeFrom:(samplingTime):timeTo;
            if(strcmpi(convertToLocal,'local'))
                [localTimeAxis,uniqueIndices,time_axis_interp] = unixUtc2localMatlab(time);
            else
                time_axis_interp = unix2matlab(time);
            end
            varTags = getVarTags(this,dbVars);
            timeAdditiveConst = 2*samplingTime;
            tic;
            clear mex;
            try
                 roughVars = MexMain({this.username, this.password, this.endpointUrl}, 1, varTags, {timeFrom-timeAdditiveConst, timeTo+timeAdditiveConst, this.varOffset, this.varCount, this.valOffset, this.valCount });
            catch
                 error('Error in tcp connect....MexMain failed...server did not response.');
            end
            count = 0;
            m = size(roughVars,1);
            t = toc;
            for c=1:m
                incr = roughVars{c};
                incr = size(incr{2},1);
                count = count + incr;
            end
            fprintf('\n');            
            msg = sprintf('All data were successfully downloaded . (%d values/%.2f sec)',count,t);           
            DisplayProcess(this.verboseMode,msg);
            
            varIds = MerbonDbVariable.empty(m,0);
            varIds(:,1) = MerbonDbVariable({});
            allVarIdx = zeros(m,1);
            isCharVariable = false(m,1);
            strMaps = repmat(struct('map',[]),m,1);
            
            %zmeneno zeros na nan
            data = nan(length(time_axis_interp),m);
            dataInCell = cell(length(time_axis_interp),m);
            dataInfo  = repmat(struct('samplePeriod',[],'numOfrecords',[]),m,1);
            
            pointerToEnd = length(dbVars)+1;
            savePosition = 0;
            
            for var = 1:m
                actVar = roughVars{var};
                varDef = actVar{1};
                varData = actVar{2};             
                timesFrom = varData(:,1);
                timesTo = varData(:,2);
                samplePeriodes = varData(:,3);
                values = varData(:,4);
                
                if(iscellstr(values) && ~isempty(values))
                    group = 'group_strings';
                    validData = true;   
                elseif(cellfun(@isinteger,values) | cellfun(@islogical,values))
                    group = 'group_singles';
                    validData = true;
                elseif(cellfun(@isfloat,values))
                    group = 'group_doubles';
                    validData = true;
                else
                    group = 'group_undefined';
                    validData = false;
                end
                
%                 validData = cellfun(@isnumeric,values);
                
                % uloz vsechny definice tak, aby se neopakovali..kdyz jich
                % najdes vice..
                varInd = this.getIndecesByIdentifiers(dbVars,{varDef});
                if(length(varInd)>1)
                    [uVI,~,ib] = intersect(allVarIdx, varInd);
                    if(isempty(uVI))
                        varInd = varInd(1);
                    else
                        varInd(ib) = [];
                        varInd = varInd(1);
                    end
                end
                if(any(allVarIdx == varInd))
                     %Jestlize uz si jednou takovouhle prom. ukladal,
                     %tak ji uloz podruhe ale uplne na konec, aby
                     %neovlivnila data, ktere jeste nejsou prirazena.
                     %typicky pri mergovani promennych.
                     allVarIdx(var) = varInd;
                     savePosition = pointerToEnd;
                     pointerToEnd = pointerToEnd + 1;
                else
                     allVarIdx(var) = varInd;
                     savePosition = varInd;
                end
               
                 
                if(validData)
                    numberOfrecords = length(values);
                    [unSamplePeriods, ia, ~] = unique(cell2mat(samplePeriodes), 'stable');
                    samplesInfo = [double(unSamplePeriods), double(unix2matlab(cell2mat(timesFrom(ia))))];
                    dataInfo(savePosition).numOfrecords = numberOfrecords;
                    dataInfo(savePosition).samplePeriod = samplesInfo;
                else
                    if(displayWarnings)
                         warning('DbVariable %d does not contain data of the same type or is empty',varInd);
                    end                    
                        % TODO: zde se prepisou data kdyz jsou dve stejny
                        % definice promeny...ale prvni ma data a druha
                        % ne..tak se to prepise nanama...neni chyba v db
                        % driveru ale v realizaci db.... nutno resit.
                        anyErrors = false;
                        varIds(savePosition) = dbVars(varInd);
                        data(:,savePosition) = nan(length(time_axis_interp),1);
                        dataInCell(:,savePosition) = num2cell(nan(length(time_axis_interp),1));
                        continue;
                end
                
                % same part for all types
                timesFrom = cell2mat(timesFrom);
                timesTo  = cell2mat(timesTo);
                samplePeriodes = cell2mat(samplePeriodes);
                
                l = size(varData,1);
                            
                d1Val = timesFrom(2:end);
                d2Val = timesTo(1:end-1);
                valDiff = d1Val - d2Val;
                validVar = valDiff>=0;
                valDiff = [valDiff;-1];
                nanInds = find(valDiff > samplePeriodes*1.1);
                nanIndsLog = valDiff < samplePeriodes*1.1;
                len = sum(nanIndsLog)*2+(length(nanIndsLog)-sum(nanIndsLog))*3;
                
                if(~validVar)
                    if(displayWarnings)
                        warning('dbVariable %d has invalid data intervals. Var will not be saved',varInd);
                    end
                    anyErrors = false;
                    varIds(savePosition) = dbVars(varInd);
                    data(:,savePosition) = nan(length(time_axis_interp),1);
                    dataInCell(:,savePosition) = num2cell(nan(length(time_axis_interp),1));
                    continue;
                end  
                
                %convert strings to their hash codes
                if(strcmp(group,'group_strings'))
                    vals = string2hash(values);
                else
                    vals = cell2mat(values);
                end
                
                % same part for all types but with hash codes if strings are presented. 
                xAxis = zeros(len,1);
                yAxis = zeros(len,1);
                
                indsFrom = 1:2:length(timesFrom)*2;
                indsTo = 2:2:(length(timesTo)*2)+1;
                zerosTime = find((timesTo-timesFrom)==0);
                timesTo(zerosTime) = timesTo(zerosTime)+1;
                xAxis(indsFrom) = timesFrom;
                xAxis(indsTo) = timesTo;
                yAxis(indsFrom) = vals;
                yAxis(indsTo) = vals;                
                nansTime = (timesFrom(nanInds+1)+timesTo(nanInds))/2;
                yAxis((2*l+1):(2*l+length(nansTime))) = NaN;
                xAxis(2*l+1:2*l+length(nansTime)) = nansTime;
                [sxAxis, IX] = sort(xAxis);
                syAxis = yAxis(IX);
                inds = [find(diff(sxAxis)>0);length(sxAxis)];
                
                if(length(inds) == 1)
                    inds = [1;2];
                elseif(isempty(inds))
                    if(displayWarnings)
                        warning('dbVariable %d does not contain minimum sample points(2). Variable will not be saved.',varInd);
                    end
                    anyErrors = false;
                    varIds(savePosition) = dbVars(varInd);
                    data(:,savePosition) = nan(length(time_axis_interp),1);
                    dataInCell(:,savePosition) = num2cell(nan(length(time_axis_interp),1));
                    continue;
                end
                
                % individual part for each group      
                switch group
                    case 'group_strings'
                    %============== STRINGS =======================================================
                        isCharVariable(savePosition) = true;
                        if(isNewestMatlab())
                            interpData = interp1(sxAxis(inds),syAxis(inds),time,'previous');
                        else
                            interpData = interp1(sxAxis(inds),syAxis(inds),time,'nearest');
                            warning('This Matlab version does not support string interpolation. If you want to get more accurate result then use 2014b or newest version');
                        end
                        if(strcmpi(convertToLocal,'local'))
                            [time, interpData] = time_shift_processing(interpData, localTimeAxis, uniqueIndices, time_axis_interp);
                        end
                        data(:,savePosition) = interpData;
                        unStrValues = unique(values);
                        if(stringRegime) % novy rezim vystupu - vrati se jen stringova mapa
                            interpDataStr = num2cell(nan(length(interpData),1)); 
                            for str=unStrValues'
                                hashOfActStr = string2hash(str);
                                indsWithThisHash = interpData == hashOfActStr; %find(interpData == hashOfActStr);
                                interpDataStr(indsWithThisHash) = str; %repcell(str{:},length(indsWithThisHash),1);
                            end
                        else
                            interpDataStr = num2cell(interpData);
                            actHashs = string2hash(unStrValues);
                            hashMap = [num2cell(actHashs), unStrValues];
                            strMaps(savePosition).map = hashMap;
                        end
                        dataInCell(:,savePosition) = interpDataStr;
                    %===========================================================================================
                    case 'group_singles'
                    %====================== SINGLES ============================================================
                        if(isNewestMatlab())
                            interpData = interp1(sxAxis(inds),syAxis(inds),time,'previous');
                        else
                            interpData = interp1(sxAxis(inds),syAxis(inds),time,'nearest');
                            warning('This Matlab version does not support single-value interpolation. If you want to get more accurate result then use 2014b or newest version');
                        end
                        if(strcmpi(convertToLocal,'local'))
                            [time, interpData] = time_shift_processing(interpData, localTimeAxis, uniqueIndices, time_axis_interp);
                        end
                        data(:,savePosition) = interpData;
                        dataInCell(:,savePosition) = num2cell(interpData);
                    %============================================================================================
                    case 'group_doubles'
                    % ===================== DOUBLES =============================================================
                        interpData = interp1(sxAxis(inds),syAxis(inds),time,interpolationMethod); 
                        if(strcmpi(convertToLocal,'local'))
                            [time, interpData] = time_shift_processing(interpData, localTimeAxis, uniqueIndices, time_axis_interp);
                        end
                        data(:,savePosition) = interpData;
                        dataInCell(:,savePosition) = num2cell(interpData);  
                    %============================================================================================
                end 
                varIds(savePosition) = dbVars(varInd);
            end %konec hlavniho cyklu pres vsechny promenne
            
            %finalni prirarazeni casu
            if(strcmpi(convertToLocal,'utc'))
                time = unix2matlab(time);
            else
                time = time_axis_interp;
            end
            dbVars = varIds;
        end
        
        function saveData(this,dbVars,timeAxis,data,valsType)
            % Save data to database. 
            %
            % INPUTS:
            % dbVars - array of MerbonDbVariable objects
            % timeAxis - row vector of Time samples in UTC
            % data - array of data if all types are only numeric, otherwise is
            % data a cell array
            % valsType  - vector of data types for specific data column. Has the
            % same length as number of columns in data matrix. N-th
            % position correspond with n*th column in data matrix. (optional)
            % valsType Options: 
            % 0 - double
            % 1 - blob
            % 2 - string
            % 3 - int64
            % 4 - not specified
            % 5 - iso date
            % 6 - bolean
            %
            % OUTPUS:
            % no output
            %
            % data is a cell matrix with n columns. Each n-th column correspond
            % with n-th variable in dbVars. Each column is a vector of any
            % type. Exaple: data = {[5x1 double],[5x1 int64],{5x1 cell}}
            % TimeAxis must have the same length as the each column. 
            % ValsType is a vecotr with the same length as the
            % number of dbVars. N-th position in valsType determine data
            % type in n-th column. 
            %
            % Options: 
            % 0 - double
            % 1 - blob
            % 2 - string
            % 3 - int64
            % 4 - not specified
            % 5 - iso date
            % 6 - bolean
            %
            % input parameter valsType is not require
            %
            % call example:
            % dbVars = MerbonDbVariable({'GroupName','Energo'});
            % timeAxis = [736884; 736885; 736886; 736887];
            % dataColumn1 = [1; 2; 3; 4];
            % dataColumn2 = [5.6; 7.77; 0; 19];
            % dataColumn3 = ['Zap'; 'VYP'; 'AUT'; 'ERROR'];
            % db.saveData(dbVars,timeAxis,[dataColumn1, dataColumn2],valsType)
            % db.saveData(dbVars,timeAxis, dataColumn1)
            % db.saveData(dbVars, timeAxis, {dataColumn1, dataColumn3})
            
            [~,n] = size(data);
            if(~iscell(data) && ismatrix(data))
%                error('4-th parameter "data" must be inserted as matrix of cells. The simplest way is to replace square bracket[] for a curly bracket{}. Input Example : data = {data1,data2,data3} instead of  data = [data1,data2,data3], where data1-3 are column vectors of any type.');
               d = cell(1,n);
               for resh=1:n
                   d(resh) = {data(:,resh)};
               end
               data = d;
            end
            forceSave = 1;
            openData = data{1};
            [l,~] = size(openData);
            data_in = cell(l,n);
            if(nargin == 5)
                varsLen = length(dbVars);
                valsLen = size(valsType,2);
                if(varsLen ~= valsLen)
                    error('Values-Type matrix must have same dimension as dbVars matrix.');
                end
            elseif(nargin<5)
                h = size(data,2);
                valsType = zeros(h,1);
                for i=1:h
                    valType = data(1,i);
                    if(iscellstr(valType{:}))
                        valsType(i) = 2;
                        data_in(:,i) = valType{:};
                    elseif(islogical(valType))
                        valsType(i) = 6;
                        data_in(:,i) = num2cell(valType{:});
                    elseif(isinteger(valType))
                        valsType(i) = 3;
                        data_in(:,i) = num2cell(valType{:});
                    elseif(isfloat(valType{:}))
                        valsType(i) = 0;
                        data_in(:,i) = num2cell(valType{:});
                    else
                        valsType(i) = 4;
                        data_in(:,i) = num2cell(valType{:});
                    end
                end
            end    
%             DisplayProcess(this.verboseMode,'',1,0);
            for var=1:length(dbVars)
                if(length(timeAxis)~=length(data{:,var}))
                     error('Time-axis matrix must have same dimension as data matrix');
                end              
                fprintf('Connected to %s DB\n',this.endpointUrl);
                fprintf('Saving %d. var\n',var);
                actVar = dbVars(var).Identifiers;
                dataTo = matlab2unix(timeAxis);
                goodTroughs = diff(dataTo);
                goodTrough = max(goodTroughs);
                if(isempty(goodTrough))
                    goodTrough = 10*60;
                end
                def_in = [num2cell(dataTo),num2cell(dataTo),num2cell(repmat(goodTrough,size(data_in,1),1)),data_in(:,var)];
                MexMain({this.username, this.password, this.endpointUrl}, 3, actVar, valsType(var), forceSave, def_in);
                dispstat('Finished.','keepprev');
            end
%             DisplayProcess(this.verboseMode,'',0,0);
        end
        
        function deleteVariable(this,var, from, to)
            % Delete selected variable from database. Allow delete only one
            % varibale.
            %
            % INPUTS:
            % var - MerbonDbVariable object
            %
            % OUTPUT:
            % not output
            %
            % call example
            % var = MerbonDbVariable({'DPGuid','CC80211D-3D29-4CC2-91A2-F69483D566B5'})
            % db.deleteVariable(var)
            if(nargin < 3)
                from = NaN;
                to = NaN;
            end
                if(size(var,1)>1 || size(var,2)>1)
                    error('Can not delete more then one variable at once');
                elseif(size(var,1)==1 && size(var,2)==1)
                     this.verboseMode = false;
                     varIds = var.Identifiers;
                     allDefsInDb = this.getAllVars({varIds(:,1:2)});
                     this.verboseMode = true;
                     if(length(allDefsInDb) > 1)
                        % if you are comenting this part of code...you act
                        % at your own risk...you can accidentally erase all the
                        % variables from DB.
                        warning('You are trying delete more than one variable from database. This var definition coresponds with %d varibales in DB. This deleting is dangerous, please use unique var definition to delete or insert password if you are sure to delete all vars.\n',length(allDefsInDb));
                        cont = input('Do you wish to continue? y/n ','s');
                        if(strcmpi(cont,'y'))
                            passwrd = input('insert password: ','s');
                            hashPassword = string2hash(passwrd);
                            if(hashPassword ~= 957835566)
                                error('Incorect password');
                            end
                        else
                            return;
                        end                   
                     end
                     DisplayProcess(this.verboseMode,'Deleting variable',0,0);
                     if(isnan(from) && isnan(to))
                        MexMain({this.username, this.password, this.endpointUrl}, 2, {varIds});
                     else
                         utcFrom = matlab2unix(from);
                         utcTo = matlab2unix(to);
                         MexMain({this.username, this.password, this.endpointUrl}, 2, {varIds}, {utcFrom, utcTo});
                     end
                     DisplayProcess(this.verboseMode,'Finished',0,0);
                else
                    error('Param "var" is not vector  with olny one element');
                end
            end
              
        function tags = getVarTags(~,dbVars)
            % separate identifiers from MerbonDbVariable
            %
            % input parameters
            %
            % dbVars = objects of MerbonDbVariable to separate
            %
            % call example
            % ids = db.GetVarsTags(dbVars)
            if(isempty(dbVars))
                tags ={};
            else
                tags = cell(length(dbVars),1);
                for var=1:length(dbVars)
                    tag = dbVars(var).Identifiers;
                    tag = tag(:,[1,2]);
                    tags(var) = {tag};
                end
            end
        end
        
        function inds = getIndecesByIdentifiers(~,vars,identifiers)
            % return indices in which have variables defined identifiers.
            % 
            % input parameters:
            %
            % vars - vector of MerbonDbVariable objects
            % identifiers - variable identifiers to be found
            %
            % variables identifier must be inserted into curly bracket.
            %
            % call example:
            % identifiers = {'StationName','ENERGO'};
            % inds = getIndecesByIdentifiers(~,vars,{identifiers})
            if(~isempty(identifiers))
                m = size(identifiers,1);
                inds = zeros(length(vars),1);
                for var=1:m
                    foundedIds = identifiers{var};                    
                    for i=1:length(vars)
                       varIDs = vars(i).Identifiers;
                       keysDef = varIDs(:,1);
                       valsDef = varIDs(:,2);    
%                        matched = zeros(1,size(varIDs,1));
                       matched = zeros(1,min(size(varIDs,1),size(foundedIds,1)));
                       matchedPosition = 1;
                       for j=1:size(foundedIds,1)%size(varIDs,1)
                           rowKey = find(strcmp(keysDef,foundedIds(j,1)));
                           rowVal = find(strcmp(valsDef,foundedIds(j,2)));
%                            rowKey = find(strcmp(keysDef(j,1),foundedIds(:,1)));
%                            rowVal = find(strcmp(valsDef(j,1),foundedIds(:,2)));
                           if(~isempty(rowKey) && ~isempty(rowVal))
                                 matched(matchedPosition) = (rowKey || rowVal);
                                 matchedPosition = matchedPosition +1;
%                                  rowMatch = (rowKey==rowVal);
%                                  matched(j) = sum(rowMatch);
                           else
%                                matched = false;
%                                break;
                                % must be checked all tags
                           end
                       end
                       if(matched)
                          inds(i) = i;
                       end
                    end
                end
                inds = inds(find(inds));
            else
                error('Identifiers array is empty');
            end
        end
        
        function [RequiredDate,vars] = getRequiredTimestamp(this,dbVars,typeOfRequest,referenceTime)
            % return vector of required dates
            %
            % input parameters:
            %
            % dbVars - vector of MerbonDBVariable objects
            % typeOfRequest - string which define type of request
            % referenceTime - matlab numeric date in UTC
            %
            % function can be called with these types of requests:
            % 'Min' - first record of the variable in the database
            % 'Max' - last record of the variable in the database - default
            % 'FisrtLess' - First record of the variable less than the reference time
            % 'FirstLessOrEqual' - First record of thevariable less or equal than the reference time
            % 'FirstGreater' - First record of the variable greater then the reference time
            % 'FirstGreaterOrEqual' - Firstrecord of the variable greater or equal than the reference time
            %
            % N-th position in RequiredDate correspond with n-th postition
            % in dbVars and vectors vars and dbVars are the same.
            % if some variable in dbVars is not unique in database, then
            % N-th position in RequiredDate correspond with n-th position
            % in vars.
            % if typeOfRequest value is equal to 'Min' or 'Max' then input
            % parameter referenceTime is not required
            % if was done request to a nonexistent variable then it will save NaN
            % value to appropriate position in RequiredDate
            %
            % example call:
            % [date,vars] = db.getRequiredTimestamp(dbVars,'FisrtLess',localMatlab2utcMatlab(referenceTime));

            allTypes = {'Min';'Max';'FirstLess';'FirstLessOrEqual';'FirstGreater';'FirstGreaterOrEqual'};
            if(nargin<3)
                warning('Input parameter "typeOfRequest" is not specified. Function will return last timestamp saved in database');
                typeOfRequest = 'max';
                referenceTime = 0;
            end
            if(~ischar(typeOfRequest))
                error('3rd input parameter must be string, which define type of request');
            end
            selectedType = find(strcmpi(allTypes,typeOfRequest));
            if(isempty(selectedType))
                error('Undefined request. As the 3rd parameter can only be inserted one of these choices: %s,%s,%s,%s,%s,%s',allTypes{1},allTypes{2},allTypes{3},allTypes{4},allTypes{5},allTypes{6});
            else
                selectedType = selectedType -1;
            end
            if(nargin<4)
                if(selectedType>1 && selectedType<6)
                    error('You have defined request type which need a reference Time, but it was not specified');
                else
                    referenceTime = 0;
                end
            end                         
            numOfVars = length(dbVars);
            if(isempty(numOfVars))
                error('No MerbonDatabase variable is specified');
            end
            DisplayProcess(this.verboseMode,'Getting required date...',0,0);
            varsDef = getVarTags(this,dbVars);
            tic;
            rougthVars = MexMain({this.username, this.password, this.endpointUrl}, 4, varsDef, { matlab2unix(referenceTime), selectedType, this.varOffset, this.varCount});
            t = toc;
            msg = sprintf('\nAll timestamps were successfully downloaded . (%d times/%.2f sec)',length(rougthVars),t);
            DisplayProcess(this.verboseMode,msg);            
            RequiredDate = nan(length(rougthVars),1);
            dispstat('','init');
            if(isempty(rougthVars))
                RequiredDate = NaN;
                return;
            else
                sortedInds = zeros(length(rougthVars),1);
                numericDate = nan(length(rougthVars),1);
                vars = MerbonDbVariable.empty(length(rougthVars),0);
            end
            for var=1:length(rougthVars)
%                 DisplayProcess(this.verboseMode,'Converting to user format',0,0,1,round(var/numOfVars)*100);
                actVar = rougthVars{var};
                actVardef = actVar(1);
                if(isempty(actVar{2}))
                    ind = this.getIndecesByIdentifiers(dbVars,actVardef);
                    sortedInds(var) = ind;
                    numericDate(var) = NaN;
                    vars(var) = MerbonDbVariable(actVardef{:});
                    continue;
                end
                actVartimestamp = actVar{2}(end,:);
                ind = this.getIndecesByIdentifiers(dbVars,actVardef);
                sortedInds(var) = ind;
                numericDate(var) = unix2matlab(double(cell2mat(actVartimestamp(2))));
                vars(var) = MerbonDbVariable(actVardef{:});
            end
%             RequiredDate(sortedInds) = numericDate;
%             vars = vars(sortedInds);
            RequiredDate = numericDate;
            
%             DisplayProcess(this.verboseMode,'Finished',0,0);
        end
        
        function [data, time, roughtData] = getDataOrigin(this, dbVar, from, to)
            % Return data exactly as are interpreted in database
            %
            % INPUTS:
            % dbVar - MerbonDbVariable object
            % from - matlab time format in utc
            % to - matlab time format in utc
            %
            % OUTPUTS:
            % data - vector of values
            % time - vector of time stamps in utc
            % roughtData - cell array of data exactly as are stored in db

            timeFrom = matlab2unix(from);
            timeTo = matlab2unix(to);
            varTags = getVarTags(this,dbVar);
            tic;
            clear mex;
            roughVars = MexMain({this.username, this.password, this.endpointUrl}, 1, varTags, {timeFrom, timeTo, this.varOffset, this.varCount, this.valOffset, this.valCount });
            if(size(roughVars,1) > 1)
                error('More than one DbVar definition was fount. Only one is alowed');
            end
            if(isempty(roughVars))
                data = [];
                time = [];
                return;
            end
            roughVars = roughVars{:};
            var_def = roughVars{1};
            var_data = roughVars{2};
            
            time = unixUtc2localMatlab(double(cell2mat(var_data(:,1)))');
            data = cell2mat(var_data(:,4));
            roughtData = var_data;
        end
        
        function [data,time,dataInCell,isVarChar,info] = getData(this,dbVars,from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings)
            % Return data matrix which has the same number of colmuns as length of
            % dbVars vector. N-th column in data matrix correspond with n-th variable in
            % dbVars. 
            % time is a vector with defined samples period.
            %
            % input parameters:
            %
            % dbVars  - objects of MerbonDbVariable to dowland
            % from - numeric matlab format time. Dowlanding will start from this time
            % to - numeric matlab format time. Dowlanding will end in this time
            % convertToLocal - If user watns to convert utc time to local
            % time, must define this parameter as "local" or "utc" if wants
            % to keep output time in UTC time. "utc" parametr is default.
            % samplingTime - sampling period in the specified time interval. In seconds
            % interpolationMethod - Method in which will be data
            % interpolated in time interval. Options:
            % 'nearest','spline','linear' - default.
            if(nargin<5)
                convertToLocal = 'utc';
                samplingTime = 5*60;
                interpolationMethod = 'linear';
                displayWarnings = true;
            end
            if(nargin<6)
                convertToLocal = 'utc';
                interpolationMethod = 'linear';
                displayWarnings = true;
            end
            if(nargin<7)
                interpolationMethod = 'linear';
                displayWarnings = true;
            end
            if(nargin<8)
                displayWarnings = true;
            end
            len = length(dbVars);
            data = [];
            info = repmat(struct('samplePeriod',[],'numOfrecords',[]),len,1);
            for var=1:len
                actVar = dbVars(var);
                [d,t,vars,dCell,errors,isVarChar,~,dataInfo] = this.getDataUnmerged(actVar,from,to,samplingTime,convertToLocal,interpolationMethod,false);
                if(errors)
                    error('This function can not finish this operation, because during the process some problems was occured. Try to call getDataUnmerged function, this function can solve your problems. See doc MerbonDatabase.getDataUnmerged');
                elseif(isempty(d))
                    if(displayWarnings)
                        warning('%d variable does not contain numerical data',var);
                    end
                    data(:,var) = nan(length(t),1);
                    dataInCell(:,var) = num2cell(nan(length(t),1));
                    continue;
                end            
                if(length(vars)== 1)
                    data(:,var) = d;
                    dataInCell(:,var) = dCell;
                    info(var) = dataInfo;
                else
                    actData = nan(length(d),1);
                    actDatacell = num2cell(nan(length(d),1));
                    for v=1:length(vars)
                        inds1 = find(~isnan(actData));
                        inds2 = find(~isnan(d(:,v)));
                        if(isempty(inds2))
                            continue;
                        end
                        [~,indsA,indsB] = intersect(inds1,inds2);
                        if(isempty(indsA) && isempty(indsB))
                            Dcut = d(inds2,v);
                            DcutCell = dCell(inds2,v);
                            actData(inds2) = Dcut;
                            actDatacell(inds2) = DcutCell;
                        else
                            Dcut = d(inds2,v);
                            DcutCell = dCell(inds2,v);
                            actData(inds2) = Dcut;
                            actDatacell(inds2) = DcutCell;
                            actData(inds1(indsA)) = NaN;
                            actDatacell(inds1(indsA)) = {NaN};
                        end 
                    end
                    data(:,var) = actData;
                    dataInCell(:,var) = actDatacell;
                    info(var) = dataInfo(v);
                end
            end
            time =t;
        end
        
       function [data,time,dataInCell,isVarChar,strMap] = getDataFast(this,dbVars,from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings,stringRegime)
           if(nargin<5)
                convertToLocal = 'utc';
                samplingTime = 5*60;
                interpolationMethod = 'linear';
                displayWarnings = false;
                stringRegime = true;
            end
            if(nargin<6)
                convertToLocal = 'utc';
                interpolationMethod = 'linear';
                displayWarnings = false;
                stringRegime = true;
            end
            if(nargin<7)
                interpolationMethod = 'linear';
                displayWarnings = false;
                stringRegime = true;
            end
            if(nargin<8)
                displayWarnings = false;
                stringRegime = true;
            end
            if(nargin < 9)
                stringRegime = true;
            end
            
            [data,time,vars,dataInCell,~,isVarChar,strMap] = this.getDataUnmerged(dbVars,from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings,stringRegime);
            if(numel(dbVars) < numel(vars))
                error(['Here is a problem with downloading of defined variables. Was returned more variables as was expected.',...
                       'Try to use the function getData or getDataUnmerged.']);
            elseif(numel(dbVars) > numel(vars))
                error('Undefined problem. See function getDataUnmerged help');
            end
       end
       
        function cached_data = getDataCached(this,dbVars,from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings)
            if(nargin<5)
                convertToLocal = 'utc';
                samplingTime = 5*60;
                interpolationMethod = 'linear';
                displayWarnings = false;
            end
            if(nargin<6)
                convertToLocal = 'utc';
                interpolationMethod = 'linear';
                displayWarnings = false;
            end
            if(nargin<7)
                interpolationMethod = 'linear';
                displayWarnings = false;
            end
            if(nargin<8)
                displayWarnings = false;
            end
            
            numVars  = numel(dbVars);
            k = 1;
            path = mfilename('fullpath');
            cached_data = repmat(struct('expression',[]),ceil(numVars/1000),1); 
            
            for i=1:1000:numVars
                if(i*k > numel(numVars))
                    KONST = numel(numVars);
                else
                    KONST = i*k;
                end
                 [data,time,vars,dataInCell,~,isVarChar] = this.getDataUnmerged(dbVars(i:KONST),from,to,samplingTime,convertToLocal,interpolationMethod,displayWarnings);
                 dataStruct.data = data;
                 dataStruct.time = time;
                 dataStruct.vars = vars;
                 dataStruct.cellData = dataInCell;
                 dataStruct.charData = isVarChar;
                 name = sprintf('data_%d_%d.mat',i, KONST);
                 save(fullfile(path,'cache',name), 'dataStruct','v-7.3');
                 cached_data(k).expression = sprintf('%s',fullfile(path,'cache',name));
                 k = k+1;
            end
            
        end
    end
    
end

